@@ -15,6 +15,8 @@ import uuid
 @ffi.callback("void(uws_res_t*, const char*, size_t, bool, void*)")
 def wsgi_on_data_handler(res, chunk, chunk_length, is_end, user_data):
     data_response = ffi.from_handle(user_data)
+    data_response.app.server.loop.is_idle = False
+
     if chunk != ffi.NULL:
         data_response.buffer.write(ffi.unpack(chunk, chunk_length))
     if bool(is_end):
@@ -29,6 +31,7 @@ def wsgi_on_data_handler(res, chunk, chunk_length, is_end, user_data):
 def wsgi_on_data_ref_abort_handler(res, user_data):
     data_retry = ffi.from_handle(user_data)
     data_retry.aborted = True
+    data_retry.server.loop.is_idle = False
     if data_retry.id is not None:
         data_retry.app._data_refs.pop(data_retry.id, None)    
     
@@ -41,7 +44,9 @@ def wsgi_on_writable_handler(res, offset, user_data):
     
     chunks = data_retry.chunks
     last_sended_offset = data_retry.last_offset
-    ssl = data_retry.app.server.SSL
+    server = data_retry.app.server
+    ssl = server.SSL
+    server.loop.is_idle = False
     content_length = data_retry.content_length
     
     data = chunks[0]
@@ -104,7 +109,6 @@ def wsgi_on_writable_handler(res, offset, user_data):
 
     return ok
 
-
 class WSGIBody:
     def __init__(self, buffer):
         self.buf = buffer
@@ -225,6 +229,7 @@ def wsgi_corked_response_start_handler(res, user_data):
 @ffi.callback("void(int, uws_res_t*, socketify_asgi_data request, void*)")
 def wsgi(ssl, response, info, user_data):
     app = ffi.from_handle(user_data)
+    app.server.loop.is_idle = False
 
     # reusing the dict is slower than cloning because we need to clear HTTP headers
     environ = dict(app.BASIC_ENVIRON)
@@ -258,9 +263,10 @@ def wsgi(ssl, response, info, user_data):
     is_chunked = False
     content_length = -1
     def write_headers(headers):
-        nonlocal headers_written, headers_set, status_text, content_length, is_chunked
+        nonlocal headers_written, headers_set, status_text, content_length, is_chunked, app
         if headers_written or not headers_set:
             return
+        app.server.loop.is_idle = False
 
         headers_written = True
 
@@ -328,7 +334,8 @@ def wsgi(ssl, response, info, user_data):
         content_length = ffi.cast("uintmax_t", content_length)
 
     def start_response(status, headers, exc_info=None):
-        nonlocal headers_set, status_text
+        nonlocal headers_set, status_text, app
+        app.server.loop.is_idle = False
         if exc_info:
             try:
                 if headers_written:
@@ -343,7 +350,8 @@ def wsgi(ssl, response, info, user_data):
         status_text = status
 
         def write(data):
-            nonlocal is_chunked
+            nonlocal is_chunked, app
+            app.server.loop.is_idle = False
             if not headers_written:
                 write_headers(headers_set)
             # will allow older frameworks only with is_chunked
@@ -372,7 +380,7 @@ def wsgi(ssl, response, info, user_data):
 
             if data_response.aborted:
                 return
-
+            data_response.app.server.loop.is_idle = False
             ssl = data_response.app.server.SSL
             data_response.environ["CONTENT_LENGTH"] = str(
                 data_response.buffer.getbuffer().nbytes
@@ -454,7 +462,11 @@ def wsgi(ssl, response, info, user_data):
         failed_chunks = None
         last_offset = -1
         data_retry = None
-        environ["wsgi.input"] = None
+        # Django do not check for None with is lame
+        # we use the same empty for everyone to avoid extra allocations
+        environ["wsgi.input"] = app.EMPTY_WSGI_BODY
+        # we also set CONTENT_LENGTH to 0 so if Django is lame again its covered
+        environ["CONTENT_LENGTH"] = "0"
         app_iter = app.wsgi(environ, start_response)
         result = None
         try:
@@ -538,6 +550,7 @@ class _WSGI:
         self.SERVER_PORT = None
         self.SERVER_WS_SCHEME = "wss" if self.server.options else "ws"
         self.wsgi = app
+        self.EMPTY_WSGI_BODY = WSGIBody(BytesIO())
         self.BASIC_ENVIRON = dict(os.environ)
         self.ws_compression = False
         self._data_refs = {}
@@ -678,10 +691,13 @@ class _WSGI:
         return self
 
     def __del__(self):
-        if self.asgi_http_info:
-            lib.socketify_destroy_asgi_app_info(self.asgi_http_info)
-        if self.asgi_ws_info:
-            lib.socketify_destroy_asgi_ws_app_info(self.asgi_ws_info)
+        try:
+            if self.asgi_http_info:
+                lib.socketify_destroy_asgi_app_info(self.asgi_http_info)
+            if self.asgi_ws_info:
+                lib.socketify_destroy_asgi_ws_app_info(self.asgi_ws_info)
+        except:
+            pass
 
 
 # "Public" WSGI interface to allow easy forks/workers
@@ -701,13 +717,27 @@ class WSGI:
         self.websocket_options = websocket_options
         self.listen_options = None
         self.task_factory_max_items = task_factory_max_items
+        self.server = None
+        self.pid_list = None
         # lifespan is not supported in WSGI
 
     def listen(self, port_or_options, handler=None):
         self.listen_options = (port_or_options, handler)
         return self
 
-    def run(self, workers=1):
+    def close(self):
+        # always wait a sec so forks can start properly if close is called too fast
+        import time
+        time.sleep(1)
+        
+        if self.server is not None:
+            self.server.close()
+        if self.pid_list is not None:
+            import signal
+            for pid in self.pid_list:
+                os.kill(pid, signal.SIGINT)
+
+    def run(self, workers=1, block=True):
         def run_task():
             server = _WSGI(
                 self.app,
@@ -719,11 +749,14 @@ class WSGI:
             if self.listen_options:
                 (port_or_options, handler) = self.listen_options
                 server.listen(port_or_options, handler)
+            self.server = server
             server.run()
 
         pid_list = []
+        
+        start = 1 if block else 0
         # fork limiting the cpu count - 1
-        for _ in range(1, workers):
+        for _ in range(start, workers):
             pid = os.fork()
             # n greater than 0 means parent process
             if not pid > 0:
@@ -731,10 +764,13 @@ class WSGI:
                 break
             pid_list.append(pid)
 
-        run_task()  # run app on the main process too :)
+        self.pid_list = pid_list
+
+        if block:
+            run_task()  # run app on the main process too :)
+            # sigint everything to gracefull shutdown
+            import signal
+            for pid in pid_list:
+                os.kill(pid, signal.SIGINT)
 
-        # sigint everything to gracefull shutdown
-        import signal
-        for pid in pid_list:
-            os.kill(pid, signal.SIGINT)
         return self
